@@ -10,7 +10,7 @@ class ParseSimpleTest(unittest.TestCase):
         self.doc = dougrain.Document.from_object({"name": "David Bowman"})
 
     def testParseSimple(self):
-        self.assertEquals(self.doc.attrs['name'], "David Bowman")
+        self.assertEquals(self.doc.properties['name'], "David Bowman")
 
     def testHasEmptyLinks(self):
         self.assertEquals(self.doc.links, {})
@@ -34,7 +34,7 @@ class ParseLinksTest(unittest.TestCase):
 
     def setUp(self):
         self.doc = dougrain.Document.from_object(
-            self.OBJECT, relative_to_url="http://localhost/wharris/dougrain")
+            self.OBJECT, base_uri="http://localhost/wharris/dougrain")
 
     def testLoadsSingleLinkHref(self):
         self.assertEquals("http://localhost/wharris/esmre",
@@ -69,7 +69,7 @@ class ParseLinksTest(unittest.TestCase):
                           [link.href for link in self.doc.links['images']])
 
     def testLinksIsNotAnAttribute(self):
-        self.assertFalse('_links' in self.doc.attrs)
+        self.assertFalse('_links' in self.doc.properties)
         self.assertFalse(hasattr(self.doc, '_links'))
 
 
@@ -99,16 +99,16 @@ class ParseEmbeddedObjectsTest(unittest.TestCase):
     def setUp(self):
         self.doc = dougrain.Document.from_object(
             self.OBJECT,
-            relative_to_url="http://localhost/people/")
+            base_uri="http://localhost/people/")
 
     def testLoadsSingleEmbeddedObject(self):
         foo = self.doc.embedded["foo"]
-        self.assertEquals("Foo", foo.attrs['name'])
-        self.assertEquals(88888888, foo.attrs['size'])
+        self.assertEquals("Foo", foo.properties['name'])
+        self.assertEquals(88888888, foo.properties['size'])
 
     def testLoadsArrayOfEmbeddedObjects(self):
         self.assertEquals(["Bar 1", "Bar 2"],
-                          [bar.attrs['title']
+                          [bar.properties['title']
                            for bar in self.doc.embedded['bar']])
 
     def testLoadsLinksInEmbeddedObject(self):
@@ -117,7 +117,7 @@ class ParseEmbeddedObjectsTest(unittest.TestCase):
         self.assertEquals("http://localhost/people/2", link.url())
 
     def testEmbeddedIsNotAnAttribute(self):
-        self.assertFalse('_embedded' in self.doc.attrs)
+        self.assertFalse('_embedded' in self.doc.properties)
         self.assertFalse(hasattr(self.doc, '_embedded'))
 
 
@@ -126,12 +126,12 @@ class CurieExpansionTest(unittest.TestCase):
         '_links': {
             'curie': [
                 {
-                    'href': "http://localhost/roles/{relation}",
+                    'href': "http://localhost/roles/{rel}",
                     'name': 'role',
                     'templated': True
                 },
                 {
-                    'href': "http://localhost/images/{relation}",
+                    'href': "http://localhost/images/{rel}",
                     'name': 'image'
                 }
             ],
@@ -143,7 +143,7 @@ class CurieExpansionTest(unittest.TestCase):
                     'curie': [
                         {
                             'href':
-                            "http://localhost/dimension/{relation}",
+                            "http://localhost/dimension/{rel}",
                             'name': 'dim',
                             'templated': True
                         }
@@ -155,7 +155,7 @@ class CurieExpansionTest(unittest.TestCase):
                     'curie': [
                         {
                             'href':
-                            "http://localhost/imagefiles/{relation}",
+                            "http://localhost/imagefiles/{rel}",
                             'name': 'image',
                             'templated': True
                         }
@@ -196,12 +196,12 @@ class RelsTest(unittest.TestCase):
         '_links': {
             'curie': [
                 {
-                    'href': "/roles/{relation}",
+                    'href': "/roles/{rel}",
                     'name': 'role',
                     'templated': True
                 },
                 {
-                    'href': "http://localhost/images/{relation}",
+                    'href': "http://localhost/images/{rel}",
                     'name': 'image',
                     'templated': True
                 }
@@ -258,7 +258,7 @@ class RelsTest(unittest.TestCase):
         self.assertTrue(consumer_role in self.doc.rels)
         consumer = self.doc.rels[consumer_role][0]
         self.assertEquals("http://localhost/clients/1", consumer.url())
-        self.assertEquals("Client 1", consumer.attrs['name'])
+        self.assertEquals("Client 1", consumer.properties['name'])
 
     def testHasApplicationLinkAndEmbeddedRels(self):
         application_role = self.doc.expand_curie('role:application')
@@ -313,26 +313,26 @@ class SerializeTests(unittest.TestCase):
 class AttributeMutationTests(unittest.TestCase):
     def testSetAttributeAddsAttribute(self):
         doc = dougrain.Document.empty()
-        doc.set_attribute('foo', "bar")
+        doc.set_property('foo', "bar")
 
-        self.assertEquals("bar", doc.attrs['foo'])
+        self.assertEquals("bar", doc.properties['foo'])
 
     def testSetAttributeUpdatesAttribute(self):
         doc = dougrain.Document.empty()
-        doc.set_attribute('foo', "bar")
+        doc.set_property('foo', "bar")
 
-        doc.set_attribute('foo', "bundy")
+        doc.set_property('foo', "bundy")
 
-        self.assertEquals("bundy", doc.attrs['foo'])
+        self.assertEquals("bundy", doc.properties['foo'])
 
     def testRemoveAttributeRemovesAttribute(self):
         doc = dougrain.Document.empty()
-        doc.set_attribute('foo', "bar")
+        doc.set_property('foo', "bar")
 
-        doc.delete_attribute('foo')
+        doc.delete_property('foo')
 
         self.assertFalse(hasattr(doc, 'foo'))
-        self.assertFalse('foo' in doc.attrs)
+        self.assertFalse('foo' in doc.properties)
 
     def testBuildObjectFromEmpty(self):
         target = {"latlng": [53.0, -0.001],
@@ -343,19 +343,15 @@ class AttributeMutationTests(unittest.TestCase):
 
         doc = dougrain.Document.empty()
 
-        for key, value in target_doc.attrs.iteritems():
-            doc.set_attribute(key, value)
+        for key, value in target_doc.properties.iteritems():
+            doc.set_property(key, value)
 
         self.assertEquals(target_doc.as_object(), doc.as_object())
 
 
-class AddLinkTests(unittest.TestCase):
-    def addLinks(self, from_doc, to_doc):
-        for rel, links in from_doc.links.iteritems():
-            if not isinstance(links, list):
-                links = [links]
-            for link in links:
-                to_doc.add_link(rel, link)
+class AddLinkStringTests(unittest.TestCase):
+    def add_link(self, doc, rel, href, **kwargs):
+        doc.add_link(rel, href, **kwargs)
 
     def testAddSimpleLink(self):
         target = {
@@ -367,7 +363,7 @@ class AddLinkTests(unittest.TestCase):
         target_doc = dougrain.Document.from_object(target, "http://localhost/")
 
         doc = dougrain.Document.empty()
-        self.addLinks(target_doc, doc)
+        self.add_link(doc, 'self', "http://localhost/1")
 
         self.assertEquals(target, doc.as_object())
         self.assertEquals(target_doc.links, doc.links)
@@ -383,7 +379,8 @@ class AddLinkTests(unittest.TestCase):
         target_doc = dougrain.Document.from_object(target, "http://localhost/")
 
         doc = dougrain.Document.empty()
-        self.addLinks(target_doc, doc)
+        self.add_link(doc, 'self', "http://localhost/1")
+        self.add_link(doc, 'child', "http://localhost/1/1")
 
         self.assertEquals(target, doc.as_object())
         self.assertEquals(target_doc.links, doc.links)
@@ -400,7 +397,9 @@ class AddLinkTests(unittest.TestCase):
         target_doc = dougrain.Document.from_object(target, "http://localhost/")
 
         doc = dougrain.Document.empty()
-        self.addLinks(target_doc, doc)
+        self.add_link(doc, 'self', "http://localhost/2")
+        self.add_link(doc, 'child', "http://localhost/2/1")
+        self.add_link(doc, 'child', "http://localhost/2/2")
 
         self.assertEquals(target, doc.as_object())
         self.assertEquals(target_doc.links, doc.links)
@@ -418,11 +417,48 @@ class AddLinkTests(unittest.TestCase):
         target_doc = dougrain.Document.from_object(target, "http://localhost/")
 
         doc = dougrain.Document.empty()
-        self.addLinks(target_doc, doc)
+        self.add_link(doc, 'self', "http://localhost/2")
+        self.add_link(doc, 'child', "http://localhost/2/1")
+        self.add_link(doc, 'child', "http://localhost/2/2")
+        self.add_link(doc, 'child', "http://localhost/2/3")
+
+        self.assertEquals(target, doc.as_object())
 
         self.assertEquals(target, doc.as_object())
         self.assertEquals(target_doc.links, doc.links)
 
+    def testAddLinkWithExtraAttributes(self):
+        target = {
+            '_links': {
+                'self': {'href': "http://localhost/2"},
+                'child': [{'href': "http://localhost/2/1",
+                           'label': "First Child"},
+                          {'href': "http://localhost/2/2",
+                           'label': "Second Child"}]
+            }
+        }
+        doc = dougrain.Document.empty()
+        self.add_link(doc, 'self', "http://localhost/2")
+        self.add_link(doc, 'child', "http://localhost/2/1",
+                      label="First Child")
+        self.add_link(doc, 'child', "http://localhost/2/2",
+                      label="Second Child")
+
+        self.assertEquals(target, doc.as_object())
+
+class AddObjectLinkTests(AddLinkStringTests):
+    def add_link(self, doc, rel, href, **kwargs):
+        link = doc.link(href, **kwargs)
+        doc.add_link(rel, link)
+
+
+class AddDocumentLinkTests(AddLinkStringTests):
+    def add_link(self, doc, rel, href, **kwargs):
+        target = dougrain.Document.empty(href)
+        self_link = target.link(href, **kwargs)
+        target.add_link('self', self_link)
+        doc.add_link(rel, target)
+
 
 class DeleteLinkTests(unittest.TestCase):
     def testDeleteOnlyLinkForRel(self):
@@ -562,6 +598,39 @@ class DeleteLinkTests(unittest.TestCase):
 
         self.assertEquals(target, doc.as_object())
 
+    def testDeleteLinkWithNoMatchingRel(self):
+        initial = {
+            '_links': {
+                'self': {'href': "http://localhost/3"},
+                'child': [{'href': "http://localhost/3/1"},
+                          {'href': "http://localhost/3/2"}],
+                'favorite': {'href': "http://localhost/3/1"}
+            }
+        }
+
+        target = {
+            '_links': {
+                'self': {'href': "http://localhost/3"},
+                'child': [{'href': "http://localhost/3/1"},
+                          {'href': "http://localhost/3/2"}],
+                'favorite': {'href': "http://localhost/3/1"}
+            }
+        }
+
+        doc = dougrain.Document.from_object(initial, "http://localhost/")
+
+        doc.delete_link(rel="".join(doc.links.keys()))
+
+        self.assertEquals(target, doc.as_object())
+
+    def testDeleteLinkWithNoLinks(self):
+        doc = dougrain.Document.empty("http://localhost/3")
+        target = dougrain.Document.empty("http://localhost/3")
+
+        doc.delete_link()
+
+        self.assertEquals(target, doc)
+
 
 class EmbedTest(unittest.TestCase):
     def setUp(self):
@@ -613,9 +682,22 @@ class EmbedTest(unittest.TestCase):
                           self.doc.embedded['item'])
 
 
+class TestIteration(unittest.TestCase):
+    def testASingleDocumentCanBeIterated(self):
+        the_doc = dougrain.Document.empty("http://localhost/1")
+        the_doc.add_link('self', "http://localhost/1")
+
+        count = 0
+        for a_doc in the_doc:
+            count += 1
+            self.assertEquals(the_doc, a_doc)
+
+        self.assertEquals(1, count)
+
+
 def make_doc(href):
     result = dougrain.Document.empty("http://localhost")
-    result.add_link("self", result.link(href))
+    result.add_link('self', href)
     return result
 
 
@@ -707,7 +789,7 @@ class DeleteEmbeddedTests(unittest.TestCase):
         target_doc = self.doc("http://localhost/3")
         target_doc.embed('child', self.doc("http://localhost/3/2"))
 
-        doc.delete_embedded(self_href="http://localhost/3/1")
+        doc.delete_embedded(href="http://localhost/3/1")
 
         self.assertEquals(target_doc.as_object(), doc.as_object())
 
@@ -723,32 +805,56 @@ class DeleteEmbeddedTests(unittest.TestCase):
 
         self.assertEquals(target_doc.as_object(), doc.as_object())
 
+    def testDeleteEmbedWithMissingRel(self):
+        doc = self.doc("http://localhost/3")
+        doc.embed('child', self.doc("http://localhost/3/1"))
+        doc.embed('child', self.doc("http://localhost/3/2"))
+        doc.embed('favorite', self.doc("http://localhost/3/1"))
+
+        target_doc = self.doc("http://localhost/3")
+        target_doc.embed('child', self.doc("http://localhost/3/1"))
+        target_doc.embed('child', self.doc("http://localhost/3/2"))
+        target_doc.embed('favorite', self.doc("http://localhost/3/1"))
+
+        missing_rel = ''.join(doc.embedded.keys()) + '_'
+        doc.delete_embedded(missing_rel)
+
+        self.assertEquals(target_doc.as_object(), doc.as_object())
+
+    def testDeleteEmbedWithNoEmbeds(self):
+        doc = self.doc("http://localhost/3")
+        target_doc = self.doc("http://localhost/3")
+
+        doc.delete_embedded()
+
+        self.assertEquals(target_doc.as_object(), doc.as_object())
+
 
 class CurieMutationTest(unittest.TestCase):
     def testSetCurie(self):
         doc = make_doc("http://localhost/3")
-        doc.set_curie('rel', "http://localhost/rels/{relation}")
+        doc.set_curie('rel', "http://localhost/rels/{rel}")
 
-        new_doc = dougrain.Document(doc.as_object(), doc.relative_to_url)
+        new_doc = dougrain.Document(doc.as_object(), doc.base_uri)
         self.assertEquals("http://localhost/rels/foo",
                           new_doc.expand_curie("rel:foo"))
 
     def testReplaceCurie(self):
         doc = make_doc("http://localhost/3")
-        doc.set_curie('rel', "http://localhost/rels/{relation}")
-        doc.set_curie('rel', "http://localhost/RELS/{relation}.html")
+        doc.set_curie('rel', "http://localhost/rels/{rel}")
+        doc.set_curie('rel', "http://localhost/RELS/{rel}.html")
 
-        new_doc = dougrain.Document(doc.as_object(), doc.relative_to_url)
+        new_doc = dougrain.Document(doc.as_object(), doc.base_uri)
         self.assertEquals("http://localhost/RELS/foo.html",
                           new_doc.expand_curie("rel:foo"))
 
     def testDropCurie(self):
         doc = make_doc("http://localhost/3")
-        doc.set_curie('rel', "http://localhost/rels/{relation}")
-        doc.set_curie('tm', "http://www.touchmachine.com/{relation}.html")
+        doc.set_curie('rel', "http://localhost/rels/{rel}")
+        doc.set_curie('tm', "http://www.touchmachine.com/{rel}.html")
         doc.drop_curie('rel')
 
-        new_doc = dougrain.Document(doc.as_object(), doc.relative_to_url)
+        new_doc = dougrain.Document(doc.as_object(), doc.base_uri)
         self.assertEquals("rel:foo", doc.expand_curie("rel:foo"))
         self.assertEquals("http://www.touchmachine.com/index.html",
                           new_doc.expand_curie("tm:index"))
@@ -759,7 +865,7 @@ class CurieHidingTests(unittest.TestCase):
         doc = dougrain.Document({
             '_links': {
                 'curie': {
-                    'href': "http://localhost/rel/{relation}",
+                    'href': "http://localhost/rel/{rel}",
                     'name': "rel"
                 },
                 'self': {
@@ -771,6 +877,114 @@ class CurieHidingTests(unittest.TestCase):
         self.assertFalse('curie' in doc.links)
                 
 
+class LinkCanonicalizationTests(unittest.TestCase):
+    def setUp(self):
+        self.doc = dougrain.Document.empty("http://localhost/1/")
+        self.doc.set_curie("role", "/roles/{rel}")
+
+    def testFindsLinksByFullRelTypeURI(self):
+        self.doc.add_link("role:app", "/apps/1")
+        self.assertEquals(self.doc.links["role:app"],
+                          self.doc.links["http://localhost/roles/app"])
+
+    def testFindsLinksByRelTypePath(self):
+        self.doc.add_link("role:app", "/apps/1")
+        self.assertEquals(self.doc.links["role:app"],
+                          self.doc.links["/roles/app"])
+
+    def testMergesLinksAddedWithSynonymous(self):
+        self.doc.add_link("role:app", "/apps/1")
+        self.doc.add_link("/roles/app", "/apps/2")
+        self.doc.add_link("http://localhost/roles/app", "/apps/3")
+
+        self.assertEquals(["/apps/1", "/apps/2", "/apps/3"],
+                           [link.href for link in self.doc.links["role:app"]])
+
+    def testMergesLinkswhenLoading(self):
+        self.doc = dougrain.Document.from_object(
+            {
+                "_links": {
+                    "curie": [
+                        {"href": "/roles/{rel}", "name": "role"}
+                    ],
+                    "self": {"href": "/1"},
+                    "role:app": {"href": "/apps/1"},
+                    "/roles/app": {"href": "/apps/2"},
+                    "http://localhost/roles/app": {"href": "/apps/3"},
+                }
+            },
+            base_uri="http://localhost/1")
+        self.doc.set_curie("role", "/roles/{rel}")
+        
+        self.assertEquals(set(["/apps/1", "/apps/2", "/apps/3"]),
+                          set(link.href for link in self.doc.links["role:app"]))
+
+
+class EmbeddedCanonicalizationTests(unittest.TestCase):
+    def setUp(self):
+        self.doc = dougrain.Document.empty("http://localhost/1/")
+        self.doc.set_curie("role", "/roles/{rel}")
+
+    def new_doc(self, uri):
+        doc = dougrain.Document.empty(uri)
+        doc.add_link('self', uri)
+        doc.set_property("name", uri)
+        return doc
+
+    def testFindsEmbeddedResourcesByFullRelTypeURI(self):
+        self.doc.embed("role:app", self.new_doc("/apps/1"))
+        self.assertEquals(self.doc.embedded["role:app"],
+                          self.doc.embedded["http://localhost/roles/app"])
+
+    def testFindsEmbeddedResourcesByRelTypePath(self):
+        self.doc.embed("role:app", self.new_doc("/apps/1"))
+        self.assertEquals(self.doc.embedded["role:app"],
+                          self.doc.embedded["/roles/app"])
+
+    def testMergesEmbeddedResourcesAddedWithSynonymous(self):
+        self.doc.embed("role:app", self.new_doc("/apps/1"))
+        self.doc.embed("/roles/app", self.new_doc("/apps/2"))
+        self.doc.embed("http://localhost/roles/app", self.new_doc("/apps/3"))
+
+        self.assertEquals(["/apps/1", "/apps/2", "/apps/3"],
+                           [embedded.properties['name']
+                            for embedded in self.doc.embedded["role:app"]])
+
+
+class EdgeCasesTests(unittest.TestCase):
+    def testUrlOfDocumentWithMultipleSelfLinksFromFirstSelfLink(self):
+        doc = dougrain.Document.empty("http://localhost")
+        doc.add_link('self', "/1")
+        doc.add_link('self', "/2")
+        self.assertEquals("http://localhost/1", doc.url())
+
+    def testSetReservedAttributeSilentlyFails(self):
+        doc = dougrain.Document.empty("http://localhost")
+        doc.set_property('_links', {'self': {'href': "/1"}})
+        doc.set_property('_embedded', {'child': {'foo': "bar"}})
+
+        self.assertFalse('_links' in doc.properties)
+        self.assertIsNone(doc.url())
+
+        self.assertFalse('_embedded' in doc.properties)
+        self.assertFalse('child' in doc.embedded)
+
+    def testDeleteReservedAttributeSilentlyFails(self):
+        doc = dougrain.Document.empty("http://localhost")
+        doc.embed('child', dougrain.Document.from_object({'foo': "bar"},
+                                                         "http://localhost"))
+        doc.add_link('self', "/1")
+
+        with self.assertRaises(KeyError):
+            doc.delete_property('_links')
+        with self.assertRaises(KeyError):
+            doc.delete_property('_embedded')
+
+        self.assertFalse('_links' in doc.properties)
+        self.assertEquals("http://localhost/1", doc.url())
+
+        self.assertFalse('_embedded' in doc.properties)
+        self.assertEquals("bar", doc.embedded['child'].properties['foo'])
 
 
 if __name__ == '__main__':
